#! /usr/bin/env swipl -G0 -T0 -L0 -s

:- prolog_load_context(directory,Dir),asserta(user:file_search_path(gcheck, Dir)). 

:- use_module(library(ordsets)).
:- use_module(library(apply)).

% use check_file(Filename) to check a file for guardness

% startup :
%
% ./gcheck.pl 
%
% alternative startup with manual script file load:
%
% swipl -s gcheck.pl
%
% then to run all tests:
%
% ?- run_tests.
%
% or to test a specific program:
%
% check_file(guarded/guarded3.logic').
%
% or to initialise the check from command line:
%
% swipl -q -t "consult(gcheck),check_file('unguarded/unguarded1.logic')"
%


check_file(Filename):-
    load_file(Filename,Clauses),
    check_program(Clauses).

check_program(Clauses):-
    check_clauses(Clauses,Clauses,0).
    
check_clauses([],_,_).

check_clauses([Goal :- _|Clauses],Program,Pos):-
    !,
    copy_term(Goal,Goal_copy),
    not(noguard_goal(Goal_copy,Program,Pos,Goal_copy,[])),
    succ(Pos,Pos_succ),
    check_clauses(Clauses,Program,Pos_succ).    

check_clauses([_Goal|Clauses],Program,Pos):-
    succ(Pos,Pos_succ),
    check_clauses(Clauses,Program,Pos_succ).    

noguard_goal(Goal,Program,GPos,Root,Trace):-
    copy_term(Program,Program_copy),
    cmember(Goal,Program_copy,Body,Pos),
    %print('checking: '),print(Root:-Goal),nl,
    ( (Trace\==[],Pos==GPos) -> 
                (guarded(Root,Goal) ->
                    nl,print('guarded: '),print(Root:-Goal),nl,fail
                ;
                    nl,print('not guarded: '),print(Root:-Goal),nl
                )
    ;
                not(member(Pos,Trace)),
                bmember(Bgoal,Body),
                noguard_goal(Bgoal,Program,GPos,Root,[Pos|Trace])
    ).

cmember(Goal,Clauses,Body,Pos):-
    cmember(Goal,Clauses,Body,0,Pos).

cmember(Goal,[Clause|_],Body,Pos,Pos):-
    % need to fail if we unify and create cycles
    unify_with_occurs_check(Goal :- Body,Clause).

cmember(Goal,[_|Clauses],Body,Pos,Pos_new):-
    succ(Pos,Pos_succ),
    cmember(Goal,Clauses,Body,Pos_succ,Pos_new).

bmember(Batom,(Batom,_)).
bmember(Batom,(_,Body)):-
    !,
    bmember(Batom,Body).

bmember(Batom,Batom).

guarded(H,B):-
    arg(I,H,Harg),
    arg(I,B,Barg),
    guarded_arg(Harg,Barg),!.

functors(T,F,L):-
    functor(T,F,L).

functors(T,F,L):-
    nonvar(T),
    compound(_,T,Targ),
    functors(Targ,F,L).

% guard that checks for unification instead of functors
%
%guarded_arg(Harg,Barg,_,_):-
%    not(subsumes_term(Harg,Barg)).

% guard that counts total amount of functors
%
guarded_arg(Harg,Barg):-    
    collect_functors(Harg,[],HFcounts_new),
    collect_functors(Barg,[],BFcounts_new),
    lessfunccount(HFcounts_new,BFcounts_new,_Func),
    term_variables(Harg,Hvars),
    term_variables(Barg,Bvars),
    list_to_ord_set(Hvars,Hvarset),
    list_to_ord_set(Bvars,Bvarset),
    ord_subset(Bvarset,Hvarset).
    %guarded_vars(Harg,Barg,Func).


guarded_vars(Harg,Barg,Func):-
    not(functors(Func,Barg,_)),
    not(functors(Func,Harg,_)),
    term_variables(Harg,Hvars),
    term_variables(Barg,Bvars),
    list_to_ord_set(Hvars,Hvarset),
    list_to_ord_set(Bvars,Bvarset),
    ord_subset(Bvarset,Hvarset).

guarded_vars(Harg,Barg,Func):-
    compound(Harg),
    arg(_,Harg,Hargarg),
    guarded_vars(Hargarg,Barg,Func).
    
guarded_vars(Harg,Barg,Func):-
    compound(Barg),
    arg(_,Barg,Bargarg),
    guarded_vars(Harg,Bargarg,Func).

lessfunccount(HFcounts,BFcounts,Func):-
    member((Func,HFcount),HFcounts),
    (member((Func,BFcount),BFcounts) -> HFcount > BFcount
                                            ; 
                                            true
    ).
        
collect_functors(Var, Fcounts,Fcounts):-
    var(Var),!.

collect_functors(Term,Fcounts,Fcounts_new):-
    Term=..[Functor|Args],
    incfunccount(Functor,Fcounts,Fcounts_tmp),
    collect_functors_list(Args,Fcounts_tmp,Fcounts_new).

collect_functors_list([Term|Terms],Fcounts,Fcounts_new):-
    collect_functors(Term,Fcounts,Fcounts_tmp),
    collect_functors_list(Terms,Fcounts_tmp,Fcounts_new).

collect_functors_list([],Fcounts,Fcounts).


incfunccount(F,[(F,Count)|Flist],[(F,Count_succ)|Flist]):-
    !,
    succ(Count,Count_succ).    

incfunccount(F,[Fcount|Fcounts],[Fcount|Fcounts_new]):-
    incfunccount(F,Fcounts,Fcounts_new).   

incfunccount(F,[],[(F,1)]).

load_file(File,Clauses):-
    open(File,read,Stream),
	call_cleanup(read_clauses(Stream,Clauses),
                 close(Stream)).
                         
read_clauses(Stream,[]) :-
        at_end_of_stream(Stream),!.

read_clauses(Stream,[Clause|Clauses]) :-
        read_term(Stream,Clause,[]),
        read_clauses(Stream,Clauses).

logic_files(Path,Filepath):-
    absolute_file_name(Path,AbsolutePath),
    directory_files(AbsolutePath,Files),
    member(Filename,Files),
    file_name_extension(_,'.logic',Filename),
    atom_concat(AbsolutePath, '/', AbsolutePath2),
    atom_concat(AbsolutePath2, Filename, Filepath).
    
:- begin_tests(gcheck).

test(guarded,[forall(logic_files(gcheck(guarded),File))]):-
    check_file(File).

test(unguarded,[forall(logic_files(gcheck(unguarded),File)),fail]):-
    check_file(File).

:- end_tests(gcheck).